@@ -1,35 +1,36 @@
 from __future__ import annotations
 
-import contextlib
 from functools import reduce
 from operator import and_
-from typing import TYPE_CHECKING, Any, Iterator, Mapping, Sequence
+from typing import TYPE_CHECKING, Any
 
 import duckdb
-from duckdb import FunctionExpression, StarExpression
+from duckdb import StarExpression
 
 from narwhals._duckdb.utils import (
     DeferredTimeZone,
+    F,
     col,
     evaluate_exprs,
-    generate_partition_by_sql,
     lit,
     native_to_narwhals_dtype,
+    window_expression,
 )
-from narwhals.dependencies import get_duckdb
-from narwhals.exceptions import InvalidOperationError
-from narwhals.typing import CompliantLazyFrame
-from narwhals.utils import (
+from narwhals._utils import (
     Implementation,
+    ValidateBackendVersion,
     Version,
     generate_temporary_column_name,
     not_implemented,
     parse_columns_to_drop,
-    parse_version,
-    validate_backend_version,
+    requires,
 )
+from narwhals.dependencies import get_duckdb
+from narwhals.exceptions import InvalidOperationError
+from narwhals.typing import CompliantLazyFrame
 
 if TYPE_CHECKING:
+    from collections.abc import Iterator, Mapping, Sequence
     from types import ModuleType
 
     import pandas as pd
@@ -43,14 +44,11 @@ if TYPE_CHECKING:
     from narwhals._duckdb.group_by import DuckDBGroupBy
     from narwhals._duckdb.namespace import DuckDBNamespace
     from narwhals._duckdb.series import DuckDBInterchangeSeries
+    from narwhals._utils import _LimitedContext
     from narwhals.dataframe import LazyFrame
     from narwhals.dtypes import DType
     from narwhals.stable.v1 import DataFrame as DataFrameV1
     from narwhals.typing import AsofJoinStrategy, JoinStrategy, LazyUniqueKeepStrategy
-    from narwhals.utils import _FullContext
-
-with contextlib.suppress(ImportError):  # requires duckdb>=1.3.0
-    from duckdb import SQLExpression
 
 
 class DuckDBLazyFrame(
@@ -58,7 +56,8 @@ class DuckDBLazyFrame(
         "DuckDBExpr",
         "duckdb.DuckDBPyRelation",
         "LazyFrame[duckdb.DuckDBPyRelation] | DataFrameV1[duckdb.DuckDBPyRelation]",
-    ]
+    ],
+    ValidateBackendVersion,
 ):
     _implementation = Implementation.DUCKDB
 
@@ -66,15 +65,19 @@ class DuckDBLazyFrame(
         self,
         df: duckdb.DuckDBPyRelation,
         *,
-        backend_version: tuple[int, ...],
         version: Version,
+        validate_backend_version: bool = False,
     ) -> None:
         self._native_frame: duckdb.DuckDBPyRelation = df
         self._version = version
-        self._backend_version = backend_version
         self._cached_native_schema: dict[str, DuckDBPyType] | None = None
         self._cached_columns: list[str] | None = None
-        validate_backend_version(self._implementation, self._backend_version)
+        if validate_backend_version:
+            self._validate_backend_version()
+
+    @property
+    def _backend_version(self) -> tuple[int, ...]:
+        return self._implementation._backend_version()
 
     @staticmethod
     def _is_native(obj: duckdb.DuckDBPyRelation | Any) -> TypeIs[duckdb.DuckDBPyRelation]:
@@ -82,11 +85,9 @@ class DuckDBLazyFrame(
 
     @classmethod
     def from_native(
-        cls, data: duckdb.DuckDBPyRelation, /, *, context: _FullContext
+        cls, data: duckdb.DuckDBPyRelation, /, *, context: _LimitedContext
     ) -> Self:
-        return cls(
-            data, backend_version=context._backend_version, version=context._version
-        )
+        return cls(data, version=context._version)
 
     def to_narwhals(
         self, *args: Any, **kwds: Any
@@ -114,9 +115,7 @@ class DuckDBLazyFrame(
     def __narwhals_namespace__(self) -> DuckDBNamespace:
         from narwhals._duckdb.namespace import DuckDBNamespace
 
-        return DuckDBNamespace(
-            backend_version=self._backend_version, version=self._version
-        )
+        return DuckDBNamespace(version=self._version)
 
     def get_column(self, name: str) -> DuckDBInterchangeSeries:
         from narwhals._duckdb.series import DuckDBInterchangeSeries
@@ -131,37 +130,31 @@ class DuckDBLazyFrame(
         self, backend: ModuleType | Implementation | str | None, **kwargs: Any
     ) -> CompliantDataFrameAny:
         if backend is None or backend is Implementation.PYARROW:
-            import pyarrow as pa  # ignore-banned-import
-
             from narwhals._arrow.dataframe import ArrowDataFrame
 
             return ArrowDataFrame(
                 self.native.arrow(),
-                backend_version=parse_version(pa),
+                validate_backend_version=True,
                 version=self._version,
                 validate_column_names=True,
             )
 
         if backend is Implementation.PANDAS:
-            import pandas as pd  # ignore-banned-import
-
             from narwhals._pandas_like.dataframe import PandasLikeDataFrame
 
             return PandasLikeDataFrame(
                 self.native.df(),
                 implementation=Implementation.PANDAS,
-                backend_version=parse_version(pd),
+                validate_backend_version=True,
                 version=self._version,
                 validate_column_names=True,
             )
 
         if backend is Implementation.POLARS:
-            import polars as pl  # ignore-banned-import
-
             from narwhals._polars.dataframe import PolarsDataFrame
 
             return PolarsDataFrame(
-                self.native.pl(), backend_version=parse_version(pl), version=self._version
+                self.native.pl(), validate_backend_version=True, version=self._version
             )
 
         msg = f"Unsupported `backend` value: {backend}"  # pragma: no cover
@@ -239,27 +232,17 @@ class DuckDBLazyFrame(
 
     def to_pandas(self) -> pd.DataFrame:
         # only if version is v1, keep around for backcompat
-        import pandas as pd  # ignore-banned-import()
-
-        if parse_version(pd) >= (1, 0, 0):
-            return self.native.df()
-        else:  # pragma: no cover
-            msg = f"Conversion to pandas requires 'pandas>=1.0.0', found {pd.__version__}"
-            raise NotImplementedError(msg)
+        return self.native.df()
 
     def to_arrow(self) -> pa.Table:
         # only if version is v1, keep around for backcompat
         return self.native.arrow()
 
     def _with_version(self, version: Version) -> Self:
-        return self.__class__(
-            self.native, version=version, backend_version=self._backend_version
-        )
+        return self.__class__(self.native, version=version)
 
     def _with_native(self, df: duckdb.DuckDBPyRelation) -> Self:
-        return self.__class__(
-            df, backend_version=self._backend_version, version=self._version
-        )
+        return self.__class__(df, version=self._version)
 
     def group_by(
         self, keys: Sequence[str] | Sequence[DuckDBExpr], *, drop_null_keys: bool
@@ -380,24 +363,15 @@ class DuckDBLazyFrame(
         self, subset: Sequence[str] | None, *, keep: LazyUniqueKeepStrategy
     ) -> Self:
         if subset_ := subset if keep == "any" else (subset or self.columns):
-            if self._backend_version < (1, 3):
-                msg = (
-                    "At least version 1.3 of DuckDB is required for `unique` operation\n"
-                    "with `subset` specified."
-                )
-                raise NotImplementedError(msg)
             # Sanitise input
             if error := self._check_columns_exist(subset_):
                 raise error
             idx_name = generate_temporary_column_name(8, self.columns)
             count_name = generate_temporary_column_name(8, [*self.columns, idx_name])
-            partition_by_sql = generate_partition_by_sql(*(subset_))
             name = count_name if keep == "none" else idx_name
-            idx_expr = SQLExpression(
-                f"{FunctionExpression('row_number')} over ({partition_by_sql})"
-            ).alias(idx_name)
-            count_expr = SQLExpression(
-                f"{FunctionExpression('count', StarExpression())} over ({partition_by_sql})"
+            idx_expr = window_expression(F("row_number"), subset_).alias(idx_name)
+            count_expr = window_expression(
+                F("count", StarExpression()), subset_, ()
             ).alias(count_name)
             return self._with_native(
                 self.native.select(StarExpression(), idx_expr, count_expr)
@@ -449,14 +423,12 @@ class DuckDBLazyFrame(
         rel = self.native
         original_columns = self.columns
 
-        not_null_condition = col_to_explode.isnotnull() & FunctionExpression(
-            "len", col_to_explode
-        ) > lit(0)
+        not_null_condition = col_to_explode.isnotnull() & F("len", col_to_explode) > lit(
+            0
+        )
         non_null_rel = rel.filter(not_null_condition).select(
             *(
-                FunctionExpression("unnest", col_to_explode).alias(name)
-                if name in columns
-                else name
+                F("unnest", col_to_explode).alias(name) if name in columns else name
                 for name in original_columns
             )
         )
@@ -503,10 +475,19 @@ class DuckDBLazyFrame(
             duckdb.sql(query).select(*[*index_, variable_name, value_name])
         )
 
+    @requires.backend_version((1, 3))
+    def with_row_index(self, name: str, order_by: Sequence[str]) -> Self:
+        if order_by is None:
+            msg = "Cannot pass `order_by` to `with_row_index` for DuckDB"
+            raise TypeError(msg)
+        expr = (window_expression(F("row_number"), order_by=order_by) - lit(1)).alias(
+            name
+        )
+        return self._with_native(self.native.select(expr, StarExpression()))
+
     gather_every = not_implemented.deprecated(
         "`LazyFrame.gather_every` is deprecated and will be removed in a future version."
     )
     tail = not_implemented.deprecated(
         "`LazyFrame.tail` is deprecated and will be removed in a future version."
     )
-    with_row_index = not_implemented()
